#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function, division

import itertools
import numpy as np


# standard atmosphere
p0 = 101325  # Pa
T0 =  293.15 # K

# constants
NA = 6.02214129e23 # mol -1

R =  8.2144621     # J / (mol K)
R_dryair = 287.058

g =  9.80665       # m s -2


# molar mass (kg/m3)
M = {
    'air': 28.97e-3,  # dry air
    'CO':  28.01e-3,
    'CO2': 44.01e-3,
    'NO2': 46.0055e-3,
    'NO':  30.01e-3,
}


def nox_to_no2(cm, model='Baechlin'):
    """\
    Convert NOx concentrations (ug/m3)
    to NO2 concentrations (ug/m3) using
    empirical equation.

    (During et al., 2011)
    """
    if model == 'Baechlin':
        A = 29.0
        B = 35.0
        C = 0.217
    elif model == 'Romberg':
        A = 103.0
        B = 130.0
        C = 0.005
    else:
        raise ValueError('Unknown model')

    return  A * cm / (cm + B) + C * cm


def calculate_mair_1d(p, ps, h):
    p = p[::-1]
    h = h[::-1]

    h_mid = 0.5 * (h[1:] + h[:-1])
    pp = np.interp(h, h_mid, p)

    mair = - np.diff(pp) / g

    return mair[::-1]




def calculate_mair(p, ps, h, q=None):
    """\
    Calculate mass of (dry) air (kg/m2).
    """
    if q is None:
        q = np.zeros_like(p)

    if np.ndim(p) == 4:
        p = p[0]
        h = h[0]
        ps = ps[0]

        if q is not None:
            q = q[0]

    # swap vertical axis
    p = p[::-1,:,:]
    h = h[::-1,:,:]

    if q is not None:
        q = q[::-1,:,:]

    ke, je, ie = p.shape
    mair = np.zeros((ke,je,ie))

    for i,j in itertools.product(range(ie), range(je)):
        h_mid = 0.5 * (h[1:,j,i] + h[:-1,j,i])

        pp = np.interp(h[:,j,i], h_mid, p[:,j,i])
        pp[0] = ps[j,i]

        if q is not None:
            mair[:,j,i] = - np.diff(pp) / g * (1.0 - q[:,j,i])
        else:
            mair[:,j,i] = - np.diff(pp) / g

    return mair[::-1,:,:]


def calculate_xco2(xm, mair, q=0.0):
    if np.ndim(xm) == 4:
        xm = xm[0]
    if np.ndim(q) == 4:
        q = q[0]

    xco2 = np.sum(mair * xm, axis=0) / np.sum(mair * (1.0 - q), axis=0)
    return 1e6 * xco2 * M['air'] / M['CO2']


def calculate_vcd(xm, p, T, h, gas):
    if np.ndim(xm) == 4:
        xm = xm[0]
        p = p[0]
        T = T[0]
        h = h[0]

    top = 20 # troposphere (TODO)

    dz = -np.diff(h[top:], axis=0)
    cm = xm_to_cm(xm[top:], p[top:], T[top:])

    if gas == 'NO2':
        cm = nox_to_no2(cm)

    cn = cm_to_cn(cm, gas)

    # NO2 VCDs (molecules per cm2)
    vcd = 1e-4 * np.sum(cn[:,:,:] * dz, axis=0)

    return vcd


def xm_to_cm(xm, p, T):
    """\
    Convert mass mixing ratios (kg/kg) to
    mass concentrations (kg/m3)
    """
    return xm * p / (R_dryair * T)

def cm_to_xm(cm, p, T):
    return cm * (R_dryair * T) / p


def cn_to_cm(cn, gas):
    return cn * M[gas] / NA


def cm_to_cn(cm, gas):
    return cm * NA / M[gas]


def xm_to_xv(xm, gas):
    """\
    Convert mass mixing ratio to volume mixing ratio.
    """
    return xm * M['air'] / M[gas]


def xv_to_xm(xv, gas):
    """\
    Convert volume mixing ratio to mass mixing ratio.
    """
    return xv * M[gas] / M['air']


def xv_to_cn(xv, p=p0, T=T0):
    """\
    Convert volume mixing ratio to number density
    """
    return xv * NA * p / (R_dryair * T)


def vertical_column_density(cn, dz, unit='cm2'):
    """\
    Compute vertical column density (VCD) from number density (per m³)
    and layer thickness (dz in metres).

    Returns VCD in cm2 or DU.
    """
    factor = {
        'cm2': 1e-4,
        'DU':  2.69e20
    }.get(unit, 1.0)

    return factor * cn * dz


def layer_thickness(p1, p2, T):
    """\
    Compute thickness between two pressure levels using
    hypsometic equation.
    """
    dz = R_dryair * T / g * np.log(p1 / p2)
    return dz






