#!/usr/bin/env python
# coding: utf-8

from six import string_types
from datetime import datetime, timedelta
import itertools
import os
import sys

from PIL import Image
from skimage import draw
from skimage import measure
from skimage import morphology as mo


import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndi
import scipy.stats as stats
import scipy.optimize as opt
import shapely
import pandas

from smartcarb import io
from smartcarb import map
from smartcarb import misc
from smartcarb import config as cfg


# ignore invalid floating-point operation
np.seterr(invalid='ignore')


class GroundTruth:
    def __init__(self, filename, origin, tracer='CO2', thr=0.1, cloud_thr=0.3,
            dn=10):

        # source origin
        self.lon_o, self.lat_o = misc.get_source_origin(origin)

        # time and ground pixels
        time = io.filename2time(filename)
        self.lon0, self.lat0, self.lonc, self.latc = io.read_ground_pixels(filename)

        # read signal and clouds
        signal_name = get_signal_name(tracer, origin)
        signal, clouds = io.read_fields(filename, [signal_name, 'CLCT'], time,
                correct_berlin=True)

        self.signal = signal
        self.clouds = clouds

        self.plume = find_plume_by_signal(signal, thr, origin,
                self.lon0, self.lat0, clouds, cloud_thr)

        # signal-weighted center of true plume
        self.center_lon = np.sum(signal[self.plume]*self.lon0[self.plume]) / np.sum(signal[self.plume])
        self.center_lat = np.sum(signal[self.plume]*self.lat0[self.plume]) / np.sum(signal[self.plume])


    def get_plume_size(self, cloud_thr):
        """
        Size of plume with cloud threshold applied after identifying plume with
        a possible different cloud fraction.

        Example: use 30% cloud threshold to idenfity plume but only 1% to get the
        number of suitable XCO2 observations.
        """
        return np.sum(self.plume & (self.clouds < cloud_thr))


    def iter_contour(self):
        if np.any(self.plume):
            return find_contour(self.lon0, self.lat0, self.plume)
        else:
            return []



def get_ground_truths(filename, origin, tracer, thr, cloud_thr, dn):

    # lon/lat of source origin
    lon_o, lat_o = misc.get_source_origin(origin)

    # time and ground pixels
    time = io.filename2time(filename)
    lon0, lat0, lonc, latc = io.read_ground_pixels(filename)

    # read signal and clouds
    signal_name = get_signal_name(tracer, origin)
    signal, clouds = io.read_fields(filename, [signal_name, 'CLCT'], time,
            correct_berlin=True)

    if thr is None:
        thr = {'CO2': 0.1, 'NO2': 0.2e15}[tracer]

    truths = find_plume_by_signal(signal, thr, origin, lon0, lat0, clouds, cloud_thr)

    # signal-weighted center of true plume
    lon_0, lat_0 = misc.compute_centroid(lon[truths], lat[truths], signal[truths])

    return lonc, latc, truths, lon_0, lat_0


def find_plume_by_signal(signal, thr, origin,
        lon0=None, lat0=None, clouds=None, cloud_thr=0.01,
        **kwargs):
    """\
    Find plume that has signal above threshold and is connected to location
    given by lon_o and lat_o.
    """
    if clouds is None:
        is_plume = signal > thr
    else:
        is_plume = (signal > thr) & (clouds < cloud_thr)

    labels = measure.label(is_plume, background=0)
    return find_plume_intersect_with_polygon(labels, origin, lon0, lat0)


def find_plume_intersect_with_polygon(labels, origin, lon, lat):

    lon_o, lat_o = misc.get_source_origin(origin)

    radius = 15.0 if origin == 'Berlin' else 5.0
    s = misc.get_circle(lon_o, lat_o, radius)

    numbers = []
    for i in set(labels.flatten()) - {0}:
        points = np.transpose([lon[labels==i], lat[labels==i]])
        m = shapely.geometry.MultiPoint(points)
        if m.intersects(s):
            numbers.append(i)

    if numbers:
        return  np.any([labels == i for i in numbers], axis=0)
    else:
        return np.zeros(labels.shape, bool)



def find_plume_at_location(labels, lon_o, lat_o, lon, lat, dn=10):
    """
    TODO:
        - allow for Polygon with intersect instead of dn.
    """

    try:
        s = get_area_around_source(lon_o, lat_o, lon, lat, dn)
    except ValueError:
        return np.zeros(labels.shape, bool)

    numbers = set(labels[s].flatten()) - {0}

    if numbers:
        return  np.any([labels == i for i in numbers], axis=0)
    else:
        return np.zeros(labels.shape, bool)


class ConfusionMatrix:
    def __init__(self, detection, truths):
        self.TP = np.sum(detection & truths)
        self.TN = np.sum(~detection & ~truths)
        self.FP = np.sum(detection & ~truths)
        self.FN = np.sum(~detection & truths)

        # the number of real positive cases in the data 
        self.P = self.TP + self.FN

        # true positive rate (sensitivity)
        self.TPR = self.TP / (self.TP + self.FN)

        # positive predictive value (precision)
        self.PPV = self.TP / (self.TP + self.FP)

        # F1-scoare (harmonic mean of TPR and PPV)
        self.F1 = 2 * self.TP / (2 * self.TP + self.FP + self.FN)


    def to_line(self, sep='\t', give_heading=False):
        if give_heading:
            return 'P\tTP\tTN\tFP\tFN\tTPR\tPPV\tF1\n'
        else:
            numbers = '\t'.join('%d' % v for v in [self.P, self.TP, self.TN, self.FP, self.FN])
            rates = '\t'.join('%.3f' % v for v in [self.TPR, self.PPV, self.F1])

            return numbers + '\t' + rates + '\n'



def label_plumes(d, n_min=0):
    """
    Label detected plume pixels. Regions with less than n_min are removed.
    """
    d[np.isnan(d)] = 0
    labels = measure.label(d, background=0)

    i = 1
    final = np.zeros_like(labels)

    for l in set(labels.flatten()):
        if l != 0 and np.sum(labels == l) >= n_min:
            final[labels == l] = i
            i += 1

    return final


def welch_satterthwaite_equation(s1, s2, n1, n2):
    """ Estimate degree of freedom. """
    a = (s1**2 / n1 + s2**2 / n2)**2
    b = s1**4 / (n1**2 * (n1-1)) + s2**2 / (n2**2 * (n2-1))
    return a/b


def local_stats(img, size):
    """
    footprint size: 1,5,9,13,25
    """
    if size == 5:
        footprint = np.array([
            [0, 1, 0],
            [1, 1, 1],
            [0, 1, 0]
        ])
    elif size == 13:
        footprint = np.array([
            [0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0],
            [1, 1, 1, 1, 1],
            [0, 1, 1, 1, 0],
            [0, 0, 1, 0, 0],
        ])

    elif size == 37:
        footprint = np.array([
            [0,0,1,1,1,0,0],
            [0,1,1,1,1,1,0],
            [1,1,1,1,1,1,1],
            [1,1,1,1,1,1,1],
            [1,1,1,1,1,1,1],
            [0,1,1,1,1,1,0],
            [0,0,1,1,1,0,0],
        ])

    elif size in [9,25,49,81,121]:
        size = int(np.sqrt(size))
        footprint = np.ones((size,size))
    else:
        raise ValueError('footprint size %d' % size)

    def nanmean(x):
        n = np.sum(np.isfinite(x))
        return np.nanmean(x) if n>0 else np.nan

    def nanvar(x):
        n = np.sum(np.isfinite(x))
        return np.nanvar(x, ddof=1) if n>1 else np.nan


    domain = np.any(img, axis=1)

    mean = np.full(img.shape, np.nan)
    mean[domain,:] = ndi.generic_filter(
            img[domain,:], nanmean, footprint=footprint, mode='constant', cval=np.nan
    )
    #var = np.full(img.shape, np.nan)
    #var[domain,:] = ndi.generic_filter(
    #        img[domain,:], nanvar, footprint=footprint, mode='constant', cval=np.nan
    #)

    n = np.full(img.shape, np.nan)
    n[domain,:] = ndi.generic_filter(
            np.isfinite(img).astype(int)[domain,:], np.count_nonzero,
            footprint=footprint, mode='constant', cval=0
    )

    return mean, n





def get_area_around_source(lon_o, lat_o, lon, lat, bg_area):
    # define area of background around Berlin

    i,j = misc.nearest_index(lon_o, lat_o, lon, lat)

    d = (np.sqrt((lon[i,j] - lon_o)**2 + (lat[i,j] - lat_o)**2))

    if d > 0.2: # about 20 km
        raise ValueError('source (likely) outside swath')

    imin, imax = max(0, i-bg_area), i+bg_area
    jmin, jmax = max(0, j-bg_area), j+bg_area
    bg = slice(imin, imax), slice(jmin, jmax)

    return bg


def detect(lon, lat, field, is_cloudy, footprint,
        lon_o, lat_o, source_area=1,
        p_thr=0.997, sigma=None,
        bg_type='ts', bg_values=None, bg_area=50,
        method='t-test', min_plume_size=0,
           origin=None,
    ):
    """\
    Detect plume in `field` at location (`lon_o`, `lat_o`)

    Parameters
    ''''''''''
    lon :: np.ndarray
        longitudes at ground pixel centers

    lat :: np.ndarray
        latitudes at ground pixel centers

    field :: np.ndarray
        synthetic satellite observations (with noise)

    is_cloudy :: np.ndarray (boolean)
        array with True where clouds

    footprint :: int
        size of footprint used for computing local stats

    lon_o, lat_o
        longitude and latitude of source origin

    source_area
        size of source area (in pixels)

    p_thr :: float
        detection threshold 

    sigma :: float
        single sounding precision (only used if footprint == 1 and thus
        sample variance cannot be computed from neighborhood)

    bg_type

    bg_values :: np.array or None
        field used for computing mean and variance of background
        if None use field

    bg_area:: int
        size of background area around origin (lon_o, lat_o)

    origin

    tracer

    """
    if bg_type == 'ts':

        mean_b = bg_values['mean']
        var_b = bg_values['var'] + sigma**2
        n_b = bg_values['n']

    elif bg_type in ['model', 'sat']:
        # estimate background mean and variance
        bg_slice = get_area_around_source(lon_o, lat_o, lon, lat, bg_area=bg_area)

        if bg_values is None:
            bg_values = field.copy()

        bg_values = bg_values[bg_slice]

        mean_b = np.nanmean(bg_values)
        var_b = np.nanvar(bg_values)

        n_b = np.isfinite(bg_values).sum()

    # estimate local mean and variance
    if footprint == 1:
        mean_s = field.copy()
        var_s = np.full(field.shape, var_b + sigma**2)
        n_s = np.ones_like(field)
    else:
        mean_s, n_s = local_stats(field, size=footprint)
        var_s = np.full(field.shape, var_b + sigma**2)


    # z-test
    t_values = np.full(mean_s.shape, np.nan)
    m = n_s > 0
    t_values[m] = (mean_s[m] - mean_b) / np.sqrt(var_s[m] / n_s[m] + var_b / n_b)

    # z-value
    is_hit = find_hits(t_values, p_thr)

    # label plumes
    labels = label_plumes(is_hit, min_plume_size)

    is_plume = find_plume_intersect_with_polygon(labels, origin, lon, lat)

    # some pixels in `in_plume` might be detected (as neighborhood) but cloudy
    # these pixels will be set to false, because they are not part of ground
    # truth and cannot be used for flux inversion (however, they are used to
    # label regions)
    is_plume[is_plume & is_cloudy] = False


    info = {'mean': mean_s, 'var': var_s, 'number': n_s,
            'mean_bg': mean_b, 'var_bg': var_b, 'n_bg': n_b,
            #'dof': dof,
            't_values': t_values, 'is_hit': is_hit,
            'labels': labels,
    }

    return is_plume, info


def get_signal_name(tracer, origin):

    suffix = 'BV' if origin == 'Berlin' else 'JV'

    if tracer == 'CO2':
        tracer = 'XCO2'

    return '%s_%s' % (tracer, suffix)




def plume_detection_algorithm(filename, tracer='CO2', error_std=0.7,
        cloud_thr=0.01, seed='orbit', use_epf=False, p_thr=0.99, footprint=3,
        bg_area=None, bg_type='ts',
        origin='Berlin', show=False, source_area=None,
        s7filename=None, use_s5=False,
        min_plume_size=0):
    """\
    Origin is Berlin or others (i.e. "Janschwalde" tracer)
    """
    if use_s5:
        if origin == 'Janschwalde':
            hours = [10,9]
        else:
            hours = [9,8]

        filename, s7filename = io.s7_to_s5(filename), filename

    is_sentinel_5 = 'Sentinel_5' in os.path.basename(filename)


    # lon/lat of source origin
    lon_o, lat_o = misc.get_source_origin(origin)

    # time and groud pixels
    time = io.filename2time(filename)
    lon0, lat0, lonc, latc = io.read_ground_pixels(filename)

    # read signal, clouds and total tracer field
    signal_name = get_signal_name(tracer, origin)
    signal, clouds = io.read_fields(filename, [signal_name, 'CLCT'], time,
            correct_berlin=True)

    true_field = io.read_total_tracer(filename, tracer, time, correct_berlin=True)

    # read noise
    if seed == 'orbit':
        _, orbit, _, lon_eq = io.process_filename(filename)
        np.random.seed(orbit + (lon_eq // 90 * 6000))
    elif seed is not None:
        np.random.seed(seed)

    noise = io.read_tracer_noise(filename, tracer, use_epf, error_std)

    # field is an array without mask (masked values should be set to nan)
    field = np.array(true_field + noise)
    field[true_field.mask | (clouds > cloud_thr)] = np.nan

    # background values
    if bg_type == 'sat':
        bg_values = field.copy()
    elif bg_type == 'model':
        bg_values = field - signal
    elif bg_type == 'ts':
        fn = cfg.BG_STATS_FILENAME[origin][tracer]
        d = pandas.read_csv(fn, sep='\t', parse_dates=True, index_col=0)

        bg_values = {}
        bg_values['mean'] = float(d[d.index == time.strftime('%Y-%m-%d %H:%M')]['mean_bg'])
        bg_values['var'] = float(d[d.index == time.strftime('%Y-%m-%d %H:%M')]['std_bg']**2)
        bg_values['n'] = np.inf

    else:
        raise ValueError("bg_type has to be 'sat', 'model' or 'ts' not '%s'" % bg_type)


    # source area
    if source_area is None:
        if is_sentinel_5:
            source_area = 4
        elif origin == 'Berlin':
            source_area = 10 # (20x20 km2)
        elif origin == 'Lippendorf':
            source_area = 5
        else:
            source_area = 2  # 4x4 km2)

    if tracer == 'NO2':
        if is_sentinel_5:
            error_std = 1.3e15
        elif error_std == 'low': # low noise scenario
            error_std = 1e15
        elif error_std == 'high': # high noise scenario
            error_std = 2e15

    # detect plume
    is_plume, info = detect(
            lon0, lat0, field, (clouds > cloud_thr), footprint,
            lon_o, lat_o, source_area,
            p_thr, error_std,
            bg_type, bg_values, bg_area,
            min_plume_size=min_plume_size,
        origin=origin
    )


    info.update({
        'lon0': lon0, 'lat0': lat0,
        'lonc': lonc, 'latc': latc,
        'signal': signal, 'field': true_field,
        'noise':  noise,  'clouds': clouds,
        'lon_o': lon_o, 'lat_o': lat_o
    })

    # get ground truth (NO2 plume above 0.2e15 cm-2 with cloud fractions below 30%)
    if tracer == 'NO2':
        info['truths_no2'] = GroundTruth(filename, origin, 'NO2',
                0.2e15, 0.3, source_area)
    else:
        info['truths_no2'] = None


    # get ground truth (CO2 plume above 0.05 ppm with cloud fractions below 30%)
    if is_sentinel_5:
        if s7filename is not None:
            info['truths'] = GroundTruth(s7filename, origin, 'CO2', 0.05, 0.3, source_area)
    else:
        info['truths'] = GroundTruth(filename, origin, 'CO2', 0.05, 0.3, source_area)

    # confusion matrix for detecting the NO2 or CO2 plume
    if is_sentinel_5:
        if s7filename is not None:
            new = project_detection(info['truths_no2'], info['truths'], is_plume)
            info['C'] = ConfusionMatrix(new, info['truths'].plume)
    else:
        info['C'] = ConfusionMatrix(is_plume, info['truths'].plume)

    if show:
        visualize(tracer, is_plume=is_plume, do_zoom=True, show_clouds=True,
                use_noise=True, marker='+', cloud_thr=cloud_thr, p_thr=p_thr,
                footprint=footprint, is_sentinel_5=is_sentinel_5, origin=origin,
                **info)

    return is_plume, info



def project_detection(in_, out, plume):
    new = np.zeros(out.plume.shape, bool)

    n = np.sum(plume)

    if n == 0:
        return new

    coords = np.transpose([in_.lonc[plume], in_.latc[plume]])

    pixels = shapely.geometry.Polygon(coords[:,0,:])
    for i in range(1,n):
        pixels = pixels.union(shapely.geometry.Polygon(coords[:,i,:]))


    for i,j in np.ndindex(*out.plume.shape):
        x,y = out.lon0[i,j], out.lat0[i,j]
        p = shapely.geometry.Point(x,y)
        new[i,j] =  p.intersects(pixels)

    return new






def find_hits(t_values, p_thr, dof=None):
    t_thr = stats.norm.ppf(p_thr)
    is_hit = t_values > t_thr
    return is_hit


def find_contour(lon, lat, binary):

    #binary = ndi.convolve(binary, np.ones((2,2))/4)

    for contour in measure.find_contours(binary, 0.5):
        coords = measure.approximate_polygon(contour, tolerance=1.0)

        n,_ = np.shape(coords)
        if n <= 4:
            continue

        i,j = coords[:,0].astype(int), coords[:,1].astype(int)
        yield lon[i,j], lat[i,j]





def visualize(tracer, lonc, latc, signal, field, noise, clouds, is_plume, labels,
        is_berlin=None, show_clouds=True, do_zoom=True, lon0=None, lat0=None,
        lon_o=None, lat_o=None,
        use_noise=True, cloud_thr=None, truths=None, truths_no2=None,
        p_thr=None, footprint=None, mean_bg=None,
        is_sentinel_5=False, origin=None,
        **kwargs):

    # tracer field
    if not use_noise:
        noise = 0.0

    if not show_clouds:
        clouds = None

    if clouds is not None and cloud_thr is not None:
        clouds = clouds > cloud_thr

    if tracer == 'CO2':
        if mean_bg is None:
            mean_bg = field.mean()
        fig = map.make_level2_map(
            lonc, latc, field - mean_bg + noise, vmin=-2, vmax=2,
            label='XCO$_2$ signal (ppm)', clct=clouds,
        )
    else:
        vmax = 10 if is_sentinel_5 else 20
        fig = map.make_level2_map(
            lonc, latc, (field + noise) / 1e15, vmin=0, vmax=vmax, truncate_cmap=True,
            label='NO$_2$ column (10$^{15}$ molecules per cm$^2$)', clct=clouds,
        )
    #
    ax, cax = fig.get_axes()
    map.add_city_labels(ax)

    # lines and labels
    lines = ax.plot(cfg.pollon, cfg.pollat, marker='s', mec='k', mfc='w', ls='',
            transform=map.ccrs.PlateCarree())
    labels = ['cloud fraction > %d%%' % (100*cloud_thr)]


    if origin == 'Berlin':
        if truths is not None:
            for i,(x,y) in enumerate(truths.iter_contour()):
                line = ax.plot(x,y, 'k-', transform=map.ccrs.PlateCarree())

                if i==0:
                    lines += line
                    labels += ['CO$_2$ plume']

        if truths_no2 is not None:
            for i,(x,y) in enumerate(truths_no2.iter_contour()):
                line = ax.plot(x,y, 'k--', transform=map.ccrs.PlateCarree())

                if i==0:
                    lines += line
                    labels += ['NO$_2$ plume']

    if is_plume is not None and np.sum(is_plume) > 0:
        lines += ax.plot(lon0[is_plume], lat0[is_plume], marker='o',
                mec='none', ms=1.5, color='k', ls='',
                transform=map.ccrs.PlateCarree()
        )
        labels += ['detected plume (p$_{thr}$ = %.2f, neighborhood = %d)'
                % (p_thr, footprint)]

    if do_zoom:
        if origin == 'Berlin' and truths is not None:
            rlon_0, rlat_0 = misc.rotpole2wgs(truths.center_lon, truths.center_lat,
                    cfg.pollon, cfg.pollat, inverse=True)
        else:
            rlon_0, rlat_0 = misc.rotpole2wgs(lon_o, lat_o, cfg.pollon,
                                              cfg.pollat, inverse=True)

        dy = 1.1
        dx = dy * 1.12
        xmin, xmax = rlon_0 - dx, rlon_0 + dx
        ymin, ymax = rlat_0 - dy, rlat_0 + dy

        try:
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
        except ValueError:
            pass

    if lines:
        ax.legend(lines, labels, loc="upper left", numpoints=1, markerscale=1.5,
                fontsize='small')


def show_all_ground_truths(filename, thr=0.1, cloud_thr=0.01, dn=10,
        output_path=None):

    # time and groud pixels
    time = io.filename2time(filename)
    lon, lat, lonc, latc = io.read_ground_pixels(filename)

    bv,jv,clct = io.read_fields(filename, ['XCO2_BV', 'XCO2_JV', 'CLCT'])

    names = ['Janschwalde', 'Schwarze Pumpe', 'Boxberg',
            'Lippendorf', 'Turow', 'Patnow']
    colors = [
        'grey',
        '#1f77b4', # Berlin
        '#ff7f0e', # J
        '#2ca02c', # SP
        '#d62728', # Bb
        '#9467bd', # Ld
        '#8c564b', # T
        '#e377c2', # P
        #'#7f7f7f',
        '#bcbd22',
        #'#17becf',
    ]
    labels = [
        'back-\nground',
        'Berlin',
        'Jänsch-\nwalde',
        'Schwarze\nPumpe',
        'Boxberg',
        'Lippen-\ndorf',
        'Turów',
        'Patnów',
        'ambig-\nuous'
    ]

    # init mask
    mask = np.zeros(lon.shape, dtype=int)

    # Berlin plume
    lon_o, lat_o = cfg.CITIES['Berlin']
    p = find_plume_by_signal(bv, thr, origin, lon, lat, dn=dn,
                            clouds=clct, cloud_thr=cloud_thr)
    mask[p] = 1

    # power plant plumes
    shape = mask.shape + (6,)
    plants = np.zeros(shape, dtype=int)

    for i, name in enumerate(names, 2):
        lon_o, lat_o = cfg.POINT_SOURCES[name]
        p = find_plume_by_signal(jv, thr, origin, lon, lat, dn=dn,
                            clouds=clct, cloud_thr=cloud_thr)

        plants[p,i-2] = i
        mask[p] = i

    # ambiguous plunmes
    mask[np.sum(plants > 0, axis=2) > 1] = len(names)+2

    labels2 = [labels[0]]
    for i, label in enumerate(labels[1:], 1):
        label = '%s\n[%d]' % (label, np.sum(mask == i))
        labels2.append(label)
    labels = labels2

    # nothing detected
    if not np.any((mask != 0) & (clct < cloud_thr)):
        return None

    # reverse numbering, colors and labels
    mask = -1 * mask + len(names) + 2
    colors = colors[::-1]
    labels = labels[::-1]

    # plot
    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=clct, is_discrete=True, cmap=colors,
        cb_labels=labels
    )

    ax, cax = fig.get_axes()
    map.add_city_labels(ax, size='medium')

    text = io.filename2label(filename)
    ax.text(0.0, 1.0, text, ha='left', va='top',
             bbox=dict(facecolor='white', alpha=1.0),
             transform=ax.transAxes
    )

    text = 'XCO$_2^\mathrm{thr}$ = %.2f ppm, d$_\mathrm{max}$ = %d px' % (thr, dn)
    ax.text(0.01, 0.0, text, ha='left', va='bottom', fontsize='x-small',
             transform=ax.transAxes
    )


    if output_path is not None:
        filename = os.path.basename(filename)
        filename = filename.replace('.nc', '_plumes.png')
        filename = os.path.join(output_path, filename)

        plt.savefig(filename)

    return True



def show_ground_truth(lonc, latc, signal, clouds, do_zoom=True, show_clouds=True,
        thr=0.1, cloud_thr=0.01, mask_clouds=True, lon_o=None, lat_o=None,
        dn=10, **kwargs):

    if mask_clouds:
        mask = find_plume_by_signal(signal, thr, origin,
                lonc.mean(2), latc.mean(2), dn=dn,
                clouds=clouds, cloud_thr=cloud_thr)
    else:
        mask = find_plume_by_signal(signal, thr, origin,
                lonc.mean(2), latc.mean(2), dn=dn)

    n_pixels = np.sum(mask)

    mask = (~mask).astype(int)
    if show_clouds:
        mask[clouds > cloud_thr] = 2

    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=None, is_discrete=True, cmap='Greys_r',
        cb_labels=['plume\n[%d]' % n_pixels, 'back-\nground', 'clouds']
    )
    ax, cax = fig.get_axes()
    map.add_city_labels(ax)

    if do_zoom:
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)

    text = 'XCO$_2^\mathrm{thr}$ = %.2f ppm, d$_\mathrm{max}$ = %d px' % (thr, dn)
    ax.text(0.01, 0.0, text, ha='left', va='bottom', fontsize='x-small',
             transform=ax.transAxes
    )


def show_labels(lonc, latc, labels, clouds, do_zoom=True, **kwargs):
    # labels

    if do_zoom:
        labels = labels.copy()
        mask = (latc.mean(2) < 50.4) | (latc.mean(2) > 53.0)
        labels[mask] = 0


    fig = map.make_level2_map(
        lonc, latc, labels, vmin=None, vmax=None,
        label='labels of detected plumes', clct=clouds > 0.01, is_discrete=True,
        cmap='nipy_spectral_r'
    )
    if do_zoom:
        ax, cax = fig.get_axes()
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)
        map.add_city_labels(ax)


def show_plume(lonc, latc, is_plume, clouds, do_zoom=True, **kwargs):

    mask = np.ones(is_plume.shape, int)
    mask[is_plume] = 0
    mask[clouds > 0.01] = 2

    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=None, is_discrete=True, cmap='Greys_r',
        cb_labels=['plume', 'back-\nground', 'clouds'],
    )
    if do_zoom:
        ax, cax = fig.get_axes()
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)
        map.add_city_labels(ax)




def estimate_detection_limit(filename, tracer, error_std, footprint=49, p_thr=0.99, use_s5=False):
    """
    """
    if tracer == 'CO2':
        cloud_thr = 0.01
    else:
        cloud_thr = 0.30

    is_plume, info = plume_detection_algorithm(filename, tracer, use_epf=True,
            error_std=error_std, bg_type='ts', origin='Berlin', p_thr=p_thr,
            footprint=footprint, cloud_thr=cloud_thr, show=False, use_s5=use_s5)

    # read CO2 signal
    signal, clouds = io.read_fields(filename, ['XCO2_BV', 'CLCT'],
            io.filename2time(filename), correct_berlin=True)


    # TP for CO2 plume
    truths = info['truths']
    TP = np.sum(is_plume & truths.plume & (clouds < 0.01))

    if TP < 1:
        return np.nan

    mask = truths.plume & (clouds < 0.01)
    signal = signal[mask]

    if TP > mask.sum():
        print('Warning: TP is larger than plume.')

    return signal[np.argsort(signal)][-TP:][0]



    min( s[mask][np.argsort(s[mask])[-info['C'].TP:]])


def read_results(origin, p_thr, footprint, use_best=False):


    data = {}
    products = []

    cases = [
        ('NO2', 'S5', 0.30),
        ('NO2', 'high', 0.30),
        ('NO2', 'high', 0.01),
        ('NO2', 'low', 0.30),
        ('NO2', 'low', 0.01),
        ('CO2', 1.0, None),
        ('CO2', 0.7, None),
        ('CO2', 0.5, None),
    ]
    for tracer, error_std, cloud_thr in cases:

        if use_best:
            if error_std == 'S5':
                p_thr, footprint = 0.95, 5
            elif tracer == 'NO2':
                p_thr, footprint = 0.99, 37
            else:
                p_thr, footprint = 0.99, 49

        case = 'case_p%d_n%d' % (100*p_thr, footprint)


        if tracer == 'CO2':
            product = 'CO2_%02d' % (10 * error_std)
        else:
            product = '%s_%s_cc%d' % (tracer, error_std, 100*cloud_thr)

        if (footprint > 9 or origin != 'Berlin') and error_std == 'S5':
            continue

        path = '/project/smartcarb/plume_detection/'
        filename = os.path.join(path, '%s_v2' % origin, case,
            'confusion_matrix_%s_%s_p%.2f_n%d.dat' % (product, origin, p_thr, footprint)
        )

        products.append(product)
        data[product] = pandas.read_csv(filename, sep='\t')

    return products, data


def plot_number_of_pixels(p_thr, footprint):

    case2label = {
            'NO2_S5_cc30':   r'NO$_2$ (Sentinel-5, cc < 30%)',
            'NO2_high_cc30': r'NO$_2$ (high noise, cc < 30%)',
            'NO2_high_cc1':  r'NO$_2$ (high noise, cc < 1%)',
            'NO2_low_cc30':  r'NO$_2$ (low noise, cc < 30%)',
            'NO2_low_cc1':   r'NO$_2$ (low noise, cc < 1%)',
            'CO2_10':        r'CO$_2$ (high noise, cc < 1%)',
            'CO2_07':        r'CO$_2$ (mid. noise, cc < 1%)',
            'CO2_05':        r'CO$_2$ (low noise, cc < 1%)'
    }

    products, data = read_results('Berlin', p_thr, footprint,
            use_best=p_thr=='best')

    fig = plt.figure()
    ax = fig.add_subplot(111)

    bins = np.arange(0,1001)
    for i, case in enumerate(sorted(data.keys())):

        if i ==0:
            P = data['CO2_05'].TP + data['CO2_05'].FN
            ax.plot(bins, [sum(P > b) for b in bins], 'k-',
                    label='real positives (cc < 30%)', zorder=5)
            print('real', sum(P>50))

        ax.plot(bins, [sum(data[case].TP > b) for b in bins], ls='-',
                label=case2label[case])

        print(case, np.mean(data[case].TP), np.median(data[case].TP))
        #print(case, sum(data[case].TP > 50), 100*sum(data[case].TP > 50) / sum(P>50))

    ax.set_xlim(0,600)
    ax.set_ylim(0,50)

    ax.set_xlabel('number of pixels (N)')
    ax.set_ylabel('number of plumes with more\nthan N detected pixels')

    if p_thr != 'best':
        ax.set_title('p$_{thr}$ = %.2f, n$_p$ = %d' % (p_thr, footprint))



    plt.xticks(np.arange(0,601,50))
    plt.yticks(np.arange(0,51,5))
    plt.grid(True)

    plt.legend(fontsize='small', ncol=1)
    plt.tight_layout()


def plot_power_plant_boxplot(origin):
    """
    Plot boxplot for power plants
    """
    p_thr = 0.99
    footprint = 5

    # indices of the 9 Level-2 data where only Janschwalde was detected for
    # all instrument scenarios
    if origin == 'Janschwalde':
        fnames = np.loadtxt('/project/smartcarb/plume_detection/Level-2_Janschwalde.dat', dtype=bytes).astype(str)

        # excl. the following orbits, because plumes overlap or detection fails
        excl_orbits = [641, 656, 846, 1010, 1230, 2132, 2296, 2322,
              2501, 3280, 3485, 3593, 3608, 4018, 4469, 4551, 5084]

        ind = [i for i,fn in enumerate(fnames) if io.filename2orbit(fn) not in excl_orbits]
        xmax = 350

    else:
        fnames = np.loadtxt('/project/smartcarb/plume_detection/Level-2_Lippendorf.dat', dtype=bytes).astype(str)

        # excl. the following orbits, because plumes overlap or detection fails
        excl_orbits = [1148, 1614, 1640, 2311, 2501, 2665, 2706, 2721, 3157, 3485,
                3500, 4059, 4074, 4238, 4525]

        ind = [i for i,fn in enumerate(fnames) if io.filename2orbit(fn) not in excl_orbits]
        xmax = 200

    print(len(fnames), len(ind))

    # read data
    products, data = read_results(origin, p_thr, footprint)

    # extract positive predictions
    X = np.transpose([[data[p].TP[i] + data[p].FP[i] for i in ind] for p in products if 'low' not in p])

    labels = []
    for p in products:

        if 'low' in p:
            continue

        n = np.sum(np.array([data[p].TP[i] + data[p].FP[i] for i in ind]) > 10)
        print(p,n)

        labels.append({
            'NO2_high_cc30': r'NO$_2$ high$_{30}$ (%d)',  # high
            'NO2_high_cc1':  r'NO$_2$ high$_1$ (%d)',     # high
            'CO2_10': r'CO$_2$ high (%d)',
            'CO2_07': r'CO$_2$ mid (%d)',
            'CO2_05': r'CO$_2$ low (%d)'}[p] % n
        )


    fig = plt.figure()
    ax = fig.add_subplot(111)

    ax.boxplot(X, whis='range', vert=False)
    ax.set_xlabel('size of detected plume')
    ax.set_xlim(-5,xmax)

    ax.set_xticks(np.arange(0,xmax+1,50))
    ax.grid(True)

    ax.set_yticklabels(labels)

    plt.tight_layout()

    return data




def plot_performance_per_footprint(case, p_thr, use_rates=True):

    if case == 'NO2_S5_cc30':
        footprints = [9,5,1]
    else:
        footprints = [121,81,49,37,25,13,9,5,1]

    if use_rates:
        rates = ['TPR', 'PPV', 'F1']
    else:
        rates = ['TP', 'FP', 'F1']

    rate2long_name = {
        'TP': 'true positive',
        'FP': 'false positive',
        'TPR': 'true-positive rate',
        'PPV': 'positive predictive value',
        'F1': 'F1 score'
    }
    data = {}
    for footprint in footprints:
        products, v = read_results('Berlin', p_thr, footprint)
        data[footprint] = v[case]

    fig = plt.figure(figsize=(10,3))

    for i,rate in enumerate(rates):
        ax = fig.add_subplot(1,3,i+1)

        # remove nans (e.g. if TP == 0)
        values = [data[p][rate] for p in footprints]
        values = [v[np.isfinite(v)] for v in values]

        ax.boxplot(values, whis=np.inf, sym='', vert=False)
        ax.set_title('(%s) %s' % ('abc'[i], rate2long_name[rate]))

        if rate in ['TPR', 'PPV', 'F1']:
            ax.set_xlim(-0.01,1.01)
        else:
            ax.set_xlim(0,1000)

        if i == 0:
            ax.set_yticklabels(footprints, rotation=0, fontsize='small')
            ax.set_ylabel('size of neighborhood')
        else:
            ax.set_yticklabels([])

    plt.tight_layout()

    return data







def plot_performance(origin, p_thr, footprint, use_rates=True):



    if use_rates:
        rates = ['TPR', 'PPV', 'F1']
    else:
        rates = ['TP', 'FP', 'F1']

    rates = ['TP', 'FP', 'F1', 'TPR', 'PPV']

    labels = [
        r'NO$_2$ S5',
        r'NO$_2$ high$_{30}$',     # high
        r'NO$_2$ high$_1$',  # high
        r'NO$_2$ low$_{30}$', # low
        r'NO$_2$ low$_1$',  # low
        r'CO$_2$ high',
        r'CO$_2$ mid',
        r'CO$_2$ low',
    ]

    rate2long_name = {
        'TP': 'true positive (TP)',
        'FP': 'false positive (FP)',
        'TPR': 'true-positive rate (TPR)',
        'PPV': 'positive predictive value (PPV)',
        'F1': 'F1 score'
    }

    products, data = read_results(origin, p_thr, footprint, use_best=p_thr=='best')


    text = '\n'.join([
        '$\mathbf{Cases}:$',
        'CO$_2$ low: $\sigma_{ref}$ = 0.5 ppm',
        'CO$_2$ mid: $\sigma_{ref}$ = 0.7 ppm',
        'CO$_2$ high: $\sigma_{ref}$ = 1.0 ppm',
        'NO$_2$ low$_1$: $\sigma_{ref}$ = 1$\\times$10$^{15}$ cm$^{-2}$, cc < 1$\%$',
        'NO$_2$ low$_{30}$: $\sigma_{ref}$ = 1$\\times$10$^{15}$ cm$^{-2}$, cc < 30$\%$',
        'NO$_2$ high$_1$: $\sigma_{ref}$ = 2$\\times$10$^{15}$ cm$^{-2}$, cc < 1$\%$',
        'NO$_2$ high$_{30}$: $\sigma_{ref}$ = 2$\\times$10$^{15}$ cm$^{-2}$, cc < 30$\%$',
        'NO$_2$ S5: Sentinel-5'
    ])


    fig = plt.figure(figsize=(10,6))
    fig.text(0.69,0.47, text, va='top', fontsize='small')

    for i,rate in enumerate(rates):
        ax = fig.add_subplot(2,3,i+1)

        # remove nans (e.g. if TP == 0)
        values = [data[p][rate] for p in products]
        values = [v[np.isfinite(v)] for v in values]


        ax.boxplot(values, whis='range', vert=False)
        #ax.set_title('(%s) %s' % ('abc'[i], rate2long_name[rate]))
        ax.set_xlabel(rate2long_name[rate])

        if rate in ['TPR', 'PPV', 'F1']:
            ax.set_xlim(-0.01,1.01)
        else:
            if footprint == 1:
                ax.set_xlim(-10,600)
            else:
                ax.set_xlim(-10,1200)

        if i in [0,3]:
            ax.set_yticklabels(labels, rotation=0, fontsize='small')
        else:
            ax.set_yticklabels([])

        if i == 1:
            if p_thr == 'best':
                ax.set_title('%s plume' % origin)
            else:
                ax.set_title('%s plume (p$_{thr}$ = %.2f, n$_p$ = %d)' % (
                    origin, p_thr, footprint))

    plt.tight_layout()

    return fig



def main(output_path, level2_path, tracer, error_std,
        origin='Berlin', p_thr=0.95, footprint=5, cloud_thr=0.01,
        use_s5=False):

    print(output_path)
    level2_filenames = np.loadtxt(
            '/project/smartcarb/plume_detection/Level-2_%s.dat' % origin,
            dtype=bytes, usecols=(0,)
        ).astype(str)


    if tracer == 'CO2':
        product = 'CO2_%02d' % (10 * error_std)
    else:
        if use_s5:
            product = 'NO2_S5_cc%d' % (100*cloud_thr)
        else:
            product = '%s_%s_cc%d' % (tracer, error_std, 100*cloud_thr)

    output_filename = 'confusion_matrix_%s_%s_p%.2f_n%d.dat' % (product, origin, p_thr, footprint)
    output_filename = os.path.join(output_path, output_filename)

    with open(output_filename, 'w') as out:

        for i, filename in enumerate(level2_filenames):
            print(filename, product)
            sys.stdout.flush()

            is_plume, info = plume_detection_algorithm(os.path.join(level2_path, filename),
                    tracer=tracer, use_epf=True, error_std=error_std,
                    bg_type='ts',
                    origin=origin,
                    p_thr=p_thr, footprint=footprint, cloud_thr=cloud_thr,
                    show=True, use_s5=use_s5
            )

            # savefig and close
            path = os.path.join(output_path, product)
            os.makedirs(path, exist_ok=True)
            plt.savefig(
                os.path.join(path, filename.replace('.nc', '_%s.png' % product))
            )
            plt.close()


            # write confusion matrix to files
            C = info['C']

            if i == 0:
                out.write(C.to_line(give_heading=True))

            out.write(C.to_line())
            out.flush()


if __name__ == '__name__':
    pass







